from typing import Any, Iterator, List, Match, Optional, Tuple

import re

from marko import block, inline, inline_parser, string_types
from marko.helpers import Source
from marko_backlinks.dto.dto import (
    Note,
    NotePath,
    NoteTitle,
    ParsedReference,
    ReferenceContext,
)
from marko_backlinks.usecases.marko_ext.exceptions import (
    TwoTitlesFoundException,
)


class Document(block.Document):
    def __init__(self, text):
        self.references: List[ParsedReference] = []
        self.source_note: Optional[Note] = None
        super().__init__(text)


class Heading(block.BlockElement):
    """Heading element: (### Hello\n)"""

    priority = 6
    pattern = re.compile(
        r" {0,3}(#{1,6})((?=\s)[^\n]*?|[^\n\S]*)(?:(?<=\s)(?<!\\)#+)?[^\n\S]*$\n?",
        flags=re.M,
    )
    inline_children = True
    override = True

    def __init__(self, match):
        self.level = len(match.group(1))
        self.children = match.group(2).strip()

    @classmethod
    def match(cls, source):
        return source.expect_re(cls.pattern)

    @classmethod
    def parse(cls, source):
        m = source.match
        if len(m.group(1)) == 1:
            if source.root.source_note:
                raise TwoTitlesFoundException()
            source.root.source_note = Note(
                note_title=m.group(2).strip(),
                note_path=NotePath(f"{m.group(2).strip()}.md"),
            )
        source.consume()
        return m


class Wikilink(inline.InlineElement):

    pattern = r"(?<!\!)\[\[([\s\S]*?)\]\]"
    priority = 5

    def __init__(self, match):
        super().__init__(match)
        self.label = match.group(1)
        self.dest = f"{self.label}.md"

    @classmethod
    def find(cls, text):  # type: (str) -> List[Match[Any]]
        """This method should return an iterable containing matches of this element."""
        if isinstance(cls.pattern, string_types):
            cls.pattern = re.compile(cls.pattern)  # type: ignore
        match_list = [match for match in cls.pattern.finditer(text)]  # type: ignore
        for match in match_list:
            wikilink = Wikilink(match)
            ref = ParsedReference(
                target_note=Note(
                    note_title=NoteTitle(wikilink.label),
                    note_path=NotePath(wikilink.dest),
                ),
                context=ReferenceContext(text),
            )
            inline._root_node.references.append(ref)
        return match_list


class Wikiimage(inline.InlineElement):
    pattern = r"\!\[\[([\s\S]*?)\]\]"
    priority = 5

    def __init__(self, match):
        super().__init__(match)
        self.image_path = match.group(1)


class Link(inline.InlineElement):
    """Link: [text](/link/destination)"""

    virtual = True
    parse_children = True
    override = True

    def __init__(self, match):  # type: (Match[Any]) -> None
        self.label, self.title, self.dest = Link.extract_label_title_and_dest(
            match
        )

    @staticmethod
    def extract_label_title_and_dest(match) -> Tuple[str, str, Optional[str]]:
        label = match.group(1)
        if (
            match.group(2)
            and match.group(2)[0] == "<"
            and match.group(2)[-1] == ">"
        ):
            dest = match.group(2)[1:-1]
        else:
            dest = match.group(2) or ""
        dest = inline.Literal.strip_backslash(dest)
        title = (
            inline.Literal.strip_backslash(match.group(3)[1:-1])
            if match.group(3)
            else None
        )
        return label, title, dest


class LinkOrEmph(inline.InlineElement):
    """This is a special element, whose parsing is done specially.
    And it produces Link or Emphasis elements.
    """

    parse_children = True
    override = True

    def __new__(cls, match):  # type: (Match[Any]) -> LinkOrEmph
        return inline.parser.inline_elements[match.etype](match)  # type: ignore

    @classmethod
    def find(cls, text):  # type: (str) -> Iterator[Match[Any]]
        match_list = inline_parser.find_links_or_emphs(text, inline._root_node)
        for match in match_list:
            if match.etype == "Link":
                link = Link(match)
                if link.dest:
                    ref = ParsedReference(
                        target_note=Note(
                            note_title=NoteTitle(link.title)
                            or NoteTitle(link.label),
                            note_path=NotePath(link.dest),
                        ),
                        context=ReferenceContext(text),
                    )
                    inline._root_node.references.append(ref)
            yield match


class BacklinkSection(block.BlockElement):
    priority = 7
    pattern = re.compile(
        r"( {0,3}#{2} Linked References\n)(^[\s\S]*$)(?=(?:^ {0,3}(#{1,6})([^\n]*?|[^\n\S]*)[^\n\S]*$\n?)|$)",
        flags=re.M,
    )
    inline_children = False

    def __init__(self, match):
        self.level = 2
        self.children = "Linked References"

    @classmethod
    def match(cls, source):
        return source.expect_re(cls.pattern)

    @classmethod
    def parse(cls, source):  # type: (Source) -> Any
        state = cls(source.match)
        source.consume()
        return state



import argparse
import glob
import os

from marko import Markdown
from marko.md_renderer import MarkdownRenderer
from marko_backlinks.adapters.references_db.factories import (
    SqlReferenceDatabaseFactory,
)
from marko_backlinks.infrastructure.db_connection import ENGINE
from marko_backlinks.interfaces import references_db
from marko_backlinks.usecases.extension import ReferencesExtension

references_db.REFERENCE_DB_FACTORY = SqlReferenceDatabaseFactory(ENGINE)
converter = Markdown(
    renderer=MarkdownRenderer, extensions=[ReferencesExtension]
)


def main(directory):
    files = {}
    for filename in glob.glob(os.path.join(directory, "*.md")):
        print(filename)
        with open(filename) as file:
            ast = converter.parse(file.read())
            files[filename] = ast
    for filename, ast in files.items():
        with open(filename, "w") as file:
            text = converter.render(ast)
            file.write(text)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "directory", type=str, help="the directory that contains all the notes"
    )
    args = parser.parse_args()
    main(directory=args.directory)



import urllib.parse
from collections import defaultdict

from marko_backlinks.interfaces import references_db


class BacklinkSectionRendererMixin:
    def __init__(self):
        super().__init__()
        self.source_note = None
        self._reference_db = references_db.REFERENCE_DB_FACTORY()

    def render_backlink_section(self, element):
        return ""

    def render_wikilink(self, element):
        return f"[{element.label}]({urllib.parse.quote(element.dest)})"

    def render_wikiimage(self, element):
        image_url = urllib.parse.quote("resources/" + element.image_path)
        return f"![No title]({image_url})"

    def render_document(self, element):
        text = self.render_children(element)  # type: ignore
        db_response = self._reference_db.get_references_that_target(
            references_db.GetReferencesThatTarget(
                note_title=element.source_note.note_title
            )
        )
        ref_dict = defaultdict(list)
        for ref in db_response.references:
            ref_dict[ref.source_note].append(ref.context)
        section_content = ""
        for source_note, contexts in ref_dict.items():
            section_content += f"  * [{source_note.note_title}]({urllib.parse.quote(source_note.note_path)})\n"
            for context in contexts:
                section_content += f"    * {context}\n"
        backlink_section = "## Linked References\n" "\n" f"{section_content}"
        return text + backlink_section



from typing import Any, NewType

from dataclasses import dataclass

NoteTitle = NewType("NoteTitle", str)
NotePath = NewType("NotePath", str)
ReferenceContext = NewType("ReferenceContext", str)


@dataclass(frozen=True)
class Note:
    note_title: NoteTitle
    note_path: NotePath


@dataclass(frozen=True)
class Reference:
    source_note: Note
    target_note: Note
    context: ReferenceContext


@dataclass(frozen=True)
class ParsedReference:
    target_note: Note
    context: ReferenceContext



from sqlalchemy import Column, ForeignKey, UniqueConstraint, types
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()  # pylint: disable=invalid-name


# mypy don't support Base as base class.
class Note(Base):  # type: ignore
    __tablename__ = "notes"

    id = Column(types.INTEGER, primary_key=True, autoincrement=True)
    note_title = Column(
        types.String, nullable=False, sqlite_on_conflict_unique="IGNORE"
    )
    note_path = Column(types.String, nullable=False)
    UniqueConstraint("note_title")

    def source_notes(self):
        return [x.source_note for x in self.lower_edges]

    def target_notes(self):
        return [x.target_note for x in self.higher_edges]


# mypy don't support Base as base class.
class Reference(Base):  # type: ignore
    __tablename__ = "references"

    id = Column(types.INTEGER, primary_key=True, autoincrement=True)
    context = Column(types.String, nullable=False)
    source_note_id = Column(types.INTEGER, ForeignKey("notes.id"))
    target_note_id = Column(types.INTEGER, ForeignKey("notes.id"))
    source_note = relationship(
        Note, primaryjoin=source_note_id == Note.id, backref="lower_edges"
    )
    target_note = relationship(
        Note, primaryjoin=target_note_id == Note.id, backref="higher_edges"
    )



from typing import Callable, Optional, Tuple, Type

from abc import ABC, abstractmethod
from dataclasses import dataclass

from marko_backlinks.dto.dto import Note, NotePath, NoteTitle, Reference

NoteId = Type[int]
ReferenceId = Type[int]


@dataclass(frozen=True)
class UpsertNoteQuery:
    note: Note


@dataclass(frozen=True)
class UpsertNoteResponse:
    note: Note
    note_id: NoteId


@dataclass(frozen=True)
class UpsertReferenceQuery:
    reference: Reference
    target_note_id: NoteId
    source_note_id: NoteId


@dataclass(frozen=True)
class UpsertReferenceResponse:
    reference: Reference
    target_note_id: NoteId
    source_note_id: NoteId
    reference_id: ReferenceId


@dataclass(frozen=True)
class GetNoteByTitleQuery:
    note_title: str


@dataclass(frozen=True)
class GetNoteResponse:
    note: Note
    note_id: NoteId


@dataclass(frozen=True)
class GetReferencesThatTarget:
    note_title: str


@dataclass(frozen=True)
class GetReferencesResponse:
    references: Tuple[Reference, ...]


class IReferenceDB(ABC):
    @abstractmethod
    def upsert_note(
        self, query: UpsertNoteQuery
    ) -> Optional[UpsertNoteResponse]:
        pass

    @abstractmethod
    def upsert_reference(
        self, query: UpsertReferenceQuery
    ) -> UpsertReferenceResponse:
        pass

    @abstractmethod
    def get_note_by_title(
        self, query: GetNoteByTitleQuery
    ) -> Optional[GetNoteResponse]:
        pass

    @abstractmethod
    def get_references_that_target(
        self, query: GetReferencesThatTarget
    ) -> GetReferencesResponse:
        pass


ReferenceDbFactory = Callable[[], IReferenceDB]
REFERENCE_DB_FACTORY: ReferenceDbFactory



from marko_backlinks.usecases.marko_ext.elements import (
    BacklinkSection,
    Document,
    Heading,
    Link,
    LinkOrEmph,
    Wikiimage,
    Wikilink,
)
from marko_backlinks.usecases.marko_ext.parser import ReferenceParser
from marko_backlinks.usecases.marko_ext.renderer import (
    BacklinkSectionRendererMixin,
)


class ReferencesExtension:
    elements = [
        Document,
        Link,
        LinkOrEmph,
        Heading,
        BacklinkSection,
        Wikilink,
        Wikiimage,
    ]
    parser_mixins = [
        ReferenceParser,
    ]
    renderer_mixins = [
        BacklinkSectionRendererMixin,
    ]



from typing import Optional


class MarkoBacklinksException(Exception):
    """Base class for all business errors in marko-backlinks.
    Args:
        code: An unique identifier for this kind of error.
    """

    def __init__(self, code: str, message: Optional[str] = None) -> None:
        self.code = code
        super().__init__(message)


class TitleNotFoundException(Exception):
    def __init__(self):
        super().__init__("title_not_found", f"Title Not Found.")


class TwoTitlesFoundException(Exception):
    def __init__(self):
        super().__init__("two_titles_found", f"Two Titles Found.")



from typing import AnyStr, List, Optional, Union

from marko import Parser, block
from marko.helpers import Source
from marko_backlinks.dto.dto import Note, ParsedReference, Reference
from marko_backlinks.interfaces import references_db
from marko_backlinks.usecases.marko_ext.exceptions import TitleNotFoundException


class ReferenceParser(Parser):
    def __init__(self):
        super().__init__()
        self._reference_db = references_db.REFERENCE_DB_FACTORY()

    def parse(
        self, source_or_text
    ):  # type: (Union[Source, AnyStr]) -> Union[List[block.BlockElement], block.BlockElement]
        ast = super().parse(source_or_text=source_or_text)
        if not isinstance(source_or_text, Source):
            self._update_db_from_ast(ast)
        return ast

    def _update_db_from_ast(self, ast):
        note: Optional[Note] = ast.source_note
        if not note:
            raise TitleNotFoundException()
        db_response = self._reference_db.upsert_note(
            references_db.UpsertNoteQuery(note=note)
        )
        parsed_references: List[ParsedReference] = ast.references
        references = [
            Reference(
                source_note=note,
                target_note=p_ref.target_note,
                context=p_ref.context,
            )
            for p_ref in parsed_references
        ]
        for reference in references:
            target_note = reference.target_note
            reference_db_response = self._reference_db.upsert_note(
                references_db.UpsertNoteQuery(note=target_note)
            )
            if reference_db_response and db_response:
                self._reference_db.upsert_reference(
                    references_db.UpsertReferenceQuery(
                        reference=reference,
                        target_note_id=reference_db_response.note_id,
                        source_note_id=db_response.note_id,
                    )
                )



from marko_backlinks.adapters.references_db.query_builders import (
    SQLiteReferenceDatabase,
)
from marko_backlinks.adapters.references_db.tables import Base
from sqlalchemy.engine import Connectable
from sqlalchemy.orm import sessionmaker


class SqlReferenceDatabaseFactory:
    _engine = None
    _session = None

    def __new__(cls, db_connection: Connectable):
        cls._engine = db_connection
        Base.metadata.create_all(cls._engine)
        Session = sessionmaker(bind=cls._engine)  # pylint: disable=invalid-name
        cls._session = Session()
        return super().__new__(cls)

    @classmethod
    def __call__(cls) -> SQLiteReferenceDatabase:
        #  SQLAlchemy engine does not allow variable permutation in the string
        # But a Session does
        return SQLiteReferenceDatabase(cls._session)



# type: ignore[attr-defined]

import glob
import os
from pathlib import Path

import typer
from marko import Markdown
from marko.md_renderer import MarkdownRenderer
from marko_backlinks import __version__
from marko_backlinks.adapters.references_db.factories import (
    SqlReferenceDatabaseFactory,
)
from marko_backlinks.infrastructure.db_connection import ENGINE
from marko_backlinks.interfaces import references_db
from marko_backlinks.usecases.extension import ReferencesExtension
from rich.console import Console

references_db.REFERENCE_DB_FACTORY = SqlReferenceDatabaseFactory(ENGINE)
converter = Markdown(
    renderer=MarkdownRenderer, extensions=[ReferencesExtension]
)


app = typer.Typer(
    name="marko-backlinks",
    help="Awesome `marko-backlinks` is a Python cli/package created with https://github.com/TezRomacH/python-package-template",
    add_completion=False,
)
console = Console()


def version_callback(value: bool):
    """Prints the version of the package."""
    if value:
        console.print(
            f"[yellow]marko-backlinks[/] version: [bold blue]{__version__}[/]"
        )
        raise typer.Exit()


@app.command(name="apply")
def main(
    directory: Path = typer.Argument(
        ...,
        help="the directory that contains all the notes",
        dir_okay=True,
        exists=True,
        file_okay=False,
        writable=True,
        readable=True,
    )
):
    files = {}
    for filename in glob.glob(os.path.join(directory, "*.md")):
        with open(filename) as file:
            ast = converter.parse(file.read())
            files[filename] = ast
    for filename, ast in files.items():
        with open(filename, "w") as file:
            text = converter.render(ast)
            file.write(text)


if __name__ == "__main__":
    app()



from typing import Optional

from marko_backlinks.adapters.references_db import tables
from marko_backlinks.dto.dto import Note, Reference
from marko_backlinks.interfaces.references_db import (
    GetNoteByTitleQuery,
    GetNoteResponse,
    GetReferencesResponse,
    GetReferencesThatTarget,
    IReferenceDB,
    UpsertNoteQuery,
    UpsertNoteResponse,
    UpsertReferenceQuery,
    UpsertReferenceResponse,
)
from sqlalchemy import insert, join, select
from sqlalchemy.engine import Connectable
from sqlalchemy.orm import aliased


class SQLiteReferenceDatabase(IReferenceDB):
    def __init__(self, db_connection: Connectable):
        self._db_connection = db_connection

    def upsert_reference(
        self, query: UpsertReferenceQuery
    ) -> UpsertReferenceResponse:
        insert_stmt = insert(tables.Reference).values(
            {
                "source_note_id": query.source_note_id,
                "target_note_id": query.target_note_id,
                "context": query.reference.context,
            }
        )
        res = self._db_connection.execute(insert_stmt)
        return UpsertReferenceResponse(
            reference=query.reference,
            target_note_id=query.target_note_id,
            source_note_id=query.source_note_id,
            reference_id=res.inserted_primary_key[0],
        )

    def upsert_note(
        self, query: UpsertNoteQuery
    ) -> Optional[UpsertNoteResponse]:
        insert_stmt = insert(tables.Note).values(
            {
                "note_title": query.note.note_title,
                "note_path": query.note.note_path,
            }
        )
        res = self._db_connection.execute(insert_stmt)
        return UpsertNoteResponse(
            note=query.note, note_id=res.inserted_primary_key[0]
        )

    def get_note_by_title(
        self, query: GetNoteByTitleQuery
    ) -> Optional[GetNoteResponse]:
        select_stmt = select([tables.Note]).where(
            tables.Note.note_title == query.note_title
        )
        res = self._db_connection.execute(select_stmt).fetchone()
        if res:
            return GetNoteResponse(
                note_id=res[0], note=Note(note_title=res[1], note_path=res[2])
            )
        else:
            return None

    def get_references_that_target(
        self, query: GetReferencesThatTarget
    ) -> GetReferencesResponse:
        TargetNotes = aliased(tables.Note, name="target_notes")
        SourceNotes = aliased(tables.Note, name="source_notes")
        stmt = (
            join(
                SourceNotes,
                tables.Reference,
                onclause=SourceNotes.id == tables.Reference.source_note_id,
            )
            .join(
                TargetNotes,
                onclause=TargetNotes.id == tables.Reference.target_note_id,
            )
            .select()
            .where(TargetNotes.note_title == query.note_title)
        )
        res = self._db_connection.execute(stmt).fetchall()
        return GetReferencesResponse(
            references=tuple(
                Reference(
                    source_note=Note(note_title=row[1], note_path=row[2]),
                    target_note=Note(note_title=row[8], note_path=row[9]),
                    context=row[4],
                )
                for row in res
            )
        )
